unit game_unit;

{
  15-puzzle game unit
  solving algorithm


  game positions
   1  2  3  4
   5  6  7  8
   9 10 11 12
  13 14 15 16

  Game[0..maxmove,0...16] : byte holds original game and subsequent moves
                  0 -  position of empty field (1..16)
                   1..16 positions hold tile number

  XGame[1..16] is game while computer searches for solution
  Xmoves[1..maxmove] : byte hold direction of empty field
  direction codes:
  right : 0; up : 1; left : 2; down : 3
}

interface

uses windows,dialogs;

const maxmove = 250;
      
type Tsearch = (ssStart,ssContinue,ssSolution);
     TSearchresult = (srTimeout,srSolution,srEnd);
     TSearchMode = (smBreadthfirst,smDepthfirst);

function DFsearch(sstate : Tsearch) : Tsearchresult;
function BFsearch(sstate : Tsearch) : Tsearchresult;
procedure setReduceRange(n1,sdepth : byte);
function RBFsearch(sstate : Tsearch; var mc : byte) : TSearchResult;
function testsolution : boolean;
procedure initgame;
procedure Randomgame(lvl : byte);  //generate puzzle with level lvl
function UpdateGame : boolean;
procedure setmasks(b : boolean);
function CanSolve : boolean;
procedure GReduce(a,b,c : byte);

var searchdepth : byte = 20;
    game : array[0..maxmove,0..16] of byte;
    Gmask : array[1..16] of boolean;//true enables solution check
    moveNr : byte;
    moveCount : byte;
    searchmode : TSearchmode = smBreadthFirst;

implementation

uses unit1;

const gamedirections : array[1..16] of byte =
      ($9,$d,$d,$c,$b,$f,$f,$e,$b,$f,$f,$e,$3,$7,$7,$6);//allowed directions per
                                                        //position
var XGame : array[1..16] of byte;
    Xfield : byte;                     //0 element position
    Xmove : array[0..maxmove] of byte; //direction of empty field
    XNr : byte;
    Xdir : byte;
    PXnr : byte;
//
    reduce1 : byte;   //game start index
    reduce2 : byte;   //game end index
    reduceRange : byte;

function GetXzeroField : byte;
var i : byte;
begin
result := 0;
 i := 1;
 while result = 0 do
  if XGame[i] = 0 then result := i else inc(i);
end;

function CanSolve : boolean;
//return true if game[0,..] can be solved
var i,j,sum : byte;
begin
 sum := 0;
 for i := 1 to 16 do
  if Game[0,i] > 0 then
   for j := i+1 to 16 do
    if Game[0,j] > 0 then
     if Game[0,j] < game[0,i] then inc(sum);
 i := Game[moveNr,0];
 i := ((i-1) shr 2) and $1;
 sum := sum + i;
 result := (sum and $1) = 1;
end;

procedure initGame;
var i,j : byte;
begin
 for i := 1 to maxmove do
  for j := 0 to 16 do game[i,j] := 0;
 setMoveInfo(0,0);
 for i := 1 to 15 do Game[0,i] := i;
 Game[0,16] := 0;
 Game[0,0] := 16;
end;

procedure setmasks(b : boolean);
var i : byte;
begin
 for i := 1 to 16 do Gmask[i] := b;
end;

procedure Randomgame(lvl : byte);
//generate puzzle with level lvl in Xgame
var rd : word;
    olddir,dir,i,rc,f,zf: byte;
    OK : boolean;
begin
 for i := 1 to 15 do Game[0,i] := i;  //set solved state
 Game[0,16] := 0;
 Game[0,0] := 16;//position of empty field
 zf := 16;
 randomize;
 rd := 0;
 rc := 0;         //random move count
 olddir := $ff;
//
 repeat
  if rc = 0 then
   begin
    rd := random($100);
    rc := 4;
   end;
  dir := rd and $3;
  rd := rd shr 2;
  dec(rc);
  OK := ((1 shl dir) and gamedirections[zf] <> 0);
  OK := OK and ((olddir xor dir) <> 2);
  if OK then
   begin
    case dir of
     0 : f := zf+1;
     1 : f := zf-4;
     2 : f := zf-1;
     3 : f := zf+4;
     else f := zf;//eliminate compiler warning
    end;//case
    Game[0,zf] := Game[0,f];
    Game[0,f] := 0;
    zf := f;
    olddir := dir;
    dec(lvl);
   end;//if OK
 until lvl = 0;
 Game[0,0] := zf;
end;

function testsolution : boolean;
//test Game[moveNr,..] for full solution, no masks
var i : byte;
begin
 result := true;
 i := 1;
 while result and (i < 16) do
  begin
   result := Game[moveNr,i] = i;
   inc(i);
  end;
end;

function testPartialSolution : boolean;
//test XGame
var i : byte;
begin
 result := true;
 i := 1;
 while result and (i < 16) do
  begin
   result := (Gmask[i]=false) or (Xgame[i]=i);
   inc(i);
  end;
end;

procedure initsearch;
//common code for search procedures
//load XGame,Xfield
//set search depth
//truncate moves after moveNr
var i : byte;
begin
 searchdepth := getSearchDepth;
 for i := 1 to 16 do XGame[i] := Game[moveNr,i];
 movecount := moveNr;
 xfield := Game[moveNr,0];
 for i := 0 to maxmove do Xmove[i] := 0;
 Xnr := 0;
 Xmove[0] := $ff;//no match with next moves
end;

function DFsearch(sstate : Tsearch) : Tsearchresult;
//(depth first) search for solution
const maxtime = 5000000;
var f : byte;
    timeout : dword;
label nextmove,nextdir,testmove,moveback,moveback1;
begin
 timeout := 0;
 if sstate = ssStart then
  begin
   initsearch;  //get Xgame, set Xfield, Xnr = 0
   if testpartialsolution then
    begin
     result := srSolution;
     exit;
    end;
  end;

nextmove:

 inc(timeout);
 if timeout = maxtime then
  begin
   result := srTimeout;
   exit;
  end;
 if Xnr >= searchdepth then goto moveback
  else begin
        inc(Xnr);      //next move
        Xdir := 0;
        goto testmove;
       end;

nextdir:

 if Xdir = 3 then goto moveback1
  else inc(xdir);

testmove:

 if ((1 shl xdir) and gamedirections[xfield]) = 0 then goto nextdir;
 if (xdir xor Xmove[Xnr-1]) = $2 then goto nextdir;
 case xdir of
  0 : f := Xfield + 1;
  1 : f := Xfield - 4;
  2 : f := Xfield - 1;
  3 : f := Xfield + 4;
  else f := 0;
  end;//case
  Xgame[Xfield] := Xgame[f];
  Xgame[f] := 0;
  Xfield := f;//h;
  Xmove[Xnr] := xdir;
  if testPartialSolution then begin
                               result := srSolution;
                               exit;
                              end
   else goto nextMove;

moveback1:

 if Xnr = 1 then
  begin
   result := srEnd;
   exit;
  end;
 dec(Xnr);

moveback:

 Xdir := Xmove[Xnr];
 case xdir of
  0 : f := Xfield-1;
  1 : f := Xfield+4;
  2 : f := Xfield+1;
  3 : f := Xfield-4;
  else f := 0;
 end;//case
 Xgame[Xfield] := Xgame[f];
 Xgame[f] := 0;
 Xfield := f;
 goto nextdir;

end;

function BFsearch(sstate : Tsearch) : Tsearchresult;
//(breadth  first) search
const maxtime = 5000000;
var f : byte;
    timeout : dword;
label Lnext,Lmove,LnextDir,Lremove;
begin
 timeout := 0;
  case sstate of
   ssContinue : goto Lmove;
   ssStart    : begin
                 initsearch;
                 if testpartialsolution then
                  begin
                   result := srSolution;
                   exit;
                  end;
                 Xnr := 1;
                 PXnr := 1;
                end;
  end;//case

Lnext :

 Xdir := 0;

Lmove :

 inc(timeout);
 if timeout = maxtime then
  begin
   result := srTimeout;
   exit;
  end;

//-- test move OK --

 if ((1 shl xdir) and gamedirections[xfield]) = 0 then goto Lnextdir;
 if (xdir xor Xmove[PXnr-1]) = $2 then goto Lnextdir;

//-- do move

 case xdir of
  0 : f := Xfield + 1;
  1 : f := Xfield - 4;
  2 : f := Xfield - 1;
  3 : f := Xfield + 4;
  else f := 0;
 end;//case
 Xgame[Xfield] := Xgame[f];
 Xgame[f] := 0;
 Xfield := f;
 Xmove[PXnr] := xdir;
//
 if Pxnr = Xnr then     // test solution
  if testPartialSolution then begin
                               result := srSolution;
                               exit;
                              end
   else goto Lremove;
//
 inc(PXnr);
 goto Lnext;

Lremove:

 case xdir of
  0 : f := Xfield-1;
  1 : f := Xfield+4;
  2 : f := Xfield+1;
  3 : f := Xfield-4;
  else f := 0;
 end;//case
 Xgame[Xfield] := Xgame[f];
 Xgame[f] := 0;
 Xfield := f;

LnextDir:

 inc(Xdir);
 if Xdir < 4 then goto Lmove;
 if PXnr = 1 then
  begin
   if Xnr = searchdepth then
    begin
     result := srEnd;
     exit;
    end
    else begin
          inc(Xnr);
          goto Lnext;
         end;
  end;
 dec(PXnr);
 Xdir := Xmove[PXnr];
 goto Lremove;
end;

//-- reduction --

procedure setReduceRange(n1,sdepth : byte);
begin
 reduce1 := n1;
 reduce2 := n1 + sdepth;
 reduceRange := sdepth;
end;

function Rcompare(nr : byte) : boolean;
//compare XGame with Game[nr,..]
//return true if compare
var i : byte;
begin
 result := true;
 i := 1;
 while result and (i <= 16) do
  if XGame[i] <> Game[nr,i] then result := false else inc(i);
end;

function RBFsearch(sstate : Tsearch; var mc : byte) : TSearchResult;
//(breadth  first) search for move reduction
//start/end/range set previously
//exit 'timeout' or 'end'
//nc : new nr of moves
const maxtime = 5000000;
var f : byte;
    timeout : dword;
label Lnext,Lmove,LnextDir,Lremove;
begin
 timeout := 0;
 case sstate of
  ssStart : begin
             for f := 1 to 16 do XGame[f] := Game[Reduce1,f];
             Xnr := 1;
             PXnr := 1;
             Xmove[0] := $ff; //no reverse of previous move
             Xfield := getXZerofield;
             goto Lnext;
            end;
  ssContinue : goto Lmove;
 end;

Lnext :

 Xdir := 0;

Lmove :

 inc(timeout);
 if timeout = maxtime then
  begin
   result := srTimeout;
   exit;
  end;

//-- test move OK --

 if ((1 shl xdir) and gamedirections[xfield]) = 0 then goto Lnextdir;
 if (xdir xor Xmove[PXnr-1]) = $2 then goto Lnextdir;

//-- do move

 case xdir of
  0 : f := Xfield + 1;
  1 : f := Xfield - 4;
  2 : f := Xfield - 1;
  3 : f := Xfield + 4;
  else f := 0;
 end;//case
 Xgame[Xfield] := Xgame[f];
 Xgame[f] := 0;
 Xfield := f;
 Xmove[PXnr] := xdir;
//
 if Pxnr = Xnr then     // test solution
  if Rcompare(Reduce2) then begin
                             mc := Pxnr;
                             result := srEnd;
                             exit;
                            end
   else goto Lremove;
//
 inc(PXnr);
 goto Lnext;

Lremove:

 case xdir of
  0 : f := Xfield-1;
  1 : f := Xfield+4;
  2 : f := Xfield+1;
  3 : f := Xfield-4;
  else f := 0;
 end;//case
 Xgame[Xfield] := Xgame[f];
 Xgame[f] := 0;
 Xfield := f;

LnextDir:

 inc(Xdir);
 if Xdir < 4 then goto Lmove;
 if PXnr = 1 then
  begin
   if Xnr = ReduceRange then
    begin
     mc := $ff;
     result := srEnd;
     exit;
    end
    else begin
          inc(Xnr);
          goto Lnext;
         end;
  end;
 dec(PXnr);
 Xdir := Xmove[PXnr];
 goto Lremove;
end;

procedure GReduce(a,b,c : byte);
//update Game for shorter search path
//a: start entry  b: stop entry of Game array c: new count
//insert Xmove[1] at a+1  Xmove[2] at a+2....  Xmove[Xnr] at a+Xnr
//shift down b+1 to a+Xnr+1 ......
//adjust movecount
var i,j,xf,df : byte;
begin
 df := 0;
 xf := Game[a,0];
 for i := 1 to c do               //insert positions from Xmove[ ]
  begin
   inc(a);
   for j := 0 to 16 do Game[a,j] := Game[a-1,j];//copy previous game state
   case Xmove[i] of
    0 : df := xf + 1;
    1 : df := xf - 4;
    2 : df := xf - 1;
    3 : df := xf + 4;
    else showmessage('error GPreduce');
   end;
   Game[a,xf] := Game[a,df];
   Game[a,df] := 0;
   Game[a,0] := df;
   xf := df;
  end;
 for i := b+1 to movecount do          //remove skipped positions
  begin
   inc(a);
   for j := 0 to 16 do Game[a,j] := Game[i,j];
  end;
 movecount := a;
end;

function UpdateGame : boolean;
//called after (partial) solved game
//insert Xmoves into game[..,..] array
var i,j,zf1,zf2 : byte;
begin
 if moveNr + XNr > maxmove then begin
                                 result := false;
                                 exit;
                                end;
 zf1 := Game[moveNr,0];
 zf2 := zf1;
 for i := 1 to 16 do XGame[i] := Game[moveNr,i];
 for i := 1 to Xnr do
  begin
   inc(moveNr);
   for j:= 1 to 16 do Game[moveNr,j] := Game[moveNr-1,j];
   case Xmove[i] of
    0 : zf2 := zf1 + 1;
    1 : zf2 := zf1 - 4;
    2 : zf2 := zf1 - 1;
    3 : zf2 := zf1 + 4;
   end;
   Game[moveNr,0] := zf2;
   Game[moveNr,zf1] := Game[moveNr,zf2];
   Game[moveNr,zf2] := 0;
   zf1 := zf2;
  end;
 result := true;
end;

end.
