unit io_unit;

{
   file format:    16 bit words
   - header
   - searchlevel, searchdepth
   - searchmode
   - moveNr,movecount
   - 16 bytes Gmask
   - movecount # of 17*4 bits in 9 bytes
   - filetrailer

   no file extension, file names are prefixed with "P15-"
}

interface

uses sysutils,dialogs;

const fileprefix = 'P15-';

var gamefilename : string;

procedure saveGame;
procedure saveGameAs;
procedure OpenGame;

implementation

uses unit1,game_unit,paint_unit;

type TB2400 = array[1..2400] of byte;

const fileheader = '15-puzzle';
      filetrailer = '15-end-of-file';

var buffer : TB2400;
    bc : word; //buffer counter
    sf : boolean; //shift flag
    f: file of TB2400;

procedure clearBuffer;
var i : word;
begin
 for i := 1 to 2400 do buffer[i] := 0;
 bc := 0;
 sf := false;
end;

//-- save file --

procedure AddQuad(q : byte);
begin
 if sf = false then begin
                     inc(bc);
                     buffer[bc] := buffer[bc] or (q and $f);
                     sf := true;
                    end
  else begin
        buffer[bc] := buffer[bc] or (q shl 4);
        sf := false;
       end;
end;

procedure addbyte(b : byte);
begin
 if sf then addQuad(0);
 inc(bc);
 buffer[bc] := b;
end;

procedure addString(s : string);
var i,lt : byte;
begin
 if sf then addQuad(0);
 lt := length(s);
 addbyte(lt);
 for i := 1 to lt do addByte(byte(s[i]));
end;

procedure addGame(nr : byte);
var i : byte;
begin
 if sf then addQuad(0);
 for i := 0 to 16 do addQuad(Game[nr,i]);
end;

procedure loadbuffer;
var i : byte;
begin
 clearBuffer;
 addstring(fileheader);                        //file header
 addbyte(moveNr);                              //moveNr
 addbyte(movecount);                           //movecount
 addbyte(form1.levelrotationBtn.position);     //levelrotation button
 addbyte(form1.depthrotationBtn.position);     //depth rotation button
 addbyte(byte(searchmode));                    //searchmode
 addbyte(0);                                   //id byte
 for i := 1 to 16 do addbyte(byte(Gmask[i]));  //Gmasks
 for i := 0 to movecount do addGame(i);        //game array
 addbyte(0);                                   //id byte
 addstring(filetrailer);                       //trailer
end;

procedure saveGame;
//cc=0: OK  1: no file  2:write error
var cc : byte;
label ex;
begin
 if length(gamefilename) = 0 then  begin
                                    cc := 1;
                                    goto ex;
                                   end;
 loadbuffer;
 try
  cc := 2;
  assignfile(f,gamefilename);
  rewrite(f);
  write(f,buffer);
  cc := 0;
 finally
  closefile(f);
 end;

ex:

 case cc of
  0 : begin
       form1.msglabel.caption := 'game saved';
       setformcaption;
      end;
  1 : form1.msglabel.Caption := 'no filename';
  2 : form1.msglabel.Caption := 'write error';
 end;//case
end;

procedure saveGameAs;
//0: Ok;  1: not executed   2: error
var s1,s2 : string;
begin
 with form1.savedialog1 do
  if execute then
   begin
    s1 := extractfilepath(filename);
    s2 := extractfilename(filename);
    if pos(fileprefix,s1) <> 1 then insert(fileprefix,s2,1);
    gamefilename := s1 + s2;
    saveGame;
   end;
end;

//-- open file --

function GetQuad : byte;
begin
 if sf then begin
             result := buffer[bc] shr 4;
             sf := false;
            end 
  else begin
        inc(bc);
        result := buffer[bc] and $f;
        sf := true;
       end;
end;

function GetByte : byte;
begin
 inc(bc);
 sf := false;
 result := buffer[bc];
end;

function getString : string;
var i : byte;
begin
 result := '';
 for i := 1 to getbyte do result := result + chr(getbyte);
end;

procedure GetGame(n : byte);
var i : byte;
begin
 if sf then getQuad;
 for i := 0 to 16 do Game[n,i] := getQuad;
 if Game[n,0] = 0 then Game[n,0] := 16;
end;

procedure OpenGame;
//cc 0: OK, 1: wrong file name 2: no file 3: format error, 4: read error
var cc,i : byte;
    s : string;
label ex;
begin
 clearBuffer;
 with form1.OpenDialog1 do
  if execute then
   begin
    s := extractfilename(filename);
    if pos(fileprefix,s) <> 1 then begin
                                    cc := 1;
                                    goto ex;
                                   end;
    if fileExists(filename) = false then begin
                                          cc := 2;
                                          goto ex;
                                         end;
    try
     cc := 4;
     assignfile(f,filename);
     reset(f);
     read(f,buffer);
     cc := 0;
    finally
     closefile(f);
    end;
   end
    else exit;
//
 if cc <> 0 then goto ex;
//
 cc := 3;//format error
 s := getstring;
 if s <> fileheader then goto ex;
//
 moveNr := getByte;
 moveCount := getByte;
 with form1 do
  begin
   levelrotationBtn.position := getByte;
   levellabel.caption := inttostr(getSearchLevel);
   depthrotationBtn.position := getByte;
   depthlabel.caption := inttostr(getSearchDepth);
   searchModeBtn.BtnDown(getbyte);
  end;
 if getByte <> 0 then goto ex;
//
 for i := 1 to 16 do Gmask[i] := boolean(getByte);
 for i := 0 to movecount do getGame(i);
 if getbyte <> 0 then goto ex;
//
 s := getstring;
 if s <> filetrailer then goto ex;
//
 cc := 0; //unpack OK

ex:

 case cc of
  0 : begin
       form1.msglabel.caption := 'game loaded';
       gamefilename := form1.OpenDialog1.FileName;
       setformcaption;
       setmoveInfo(moveNr,movecount);
       paintgame(moveNr);
      end;
  1 : form1.msglabel.caption := 'wrong filename';
  2 : form1.msglabel.caption := 'no file found';
  3 : form1.msglabel.caption := 'file format error';
  4 : form1.msglabel.caption := 'read error';
 end;//case
//
 if cc <> 0 then
  begin
   initgame;
  end;
 form1.menuBtn.BtnRelease(0);
end;

end.
